private import std.math;
private import std.string;

static const int X = 0;
static const int Y = 1;
static const int Z = 2;

struct Vec3 {
public:
    static const int DIM = 3;

public:
    // data
    union {
        struct {
            float x, y, z;
        }
        float[3] _;
    }

    // ctor
    void init(float x0, float y0, float z0) {
        x = x0;
        y = y0;
        z = z0;
    }
    static Vec3 create(float x0, float y0, float z0) {
        Vec3 r;
        r.x = x0;
        r.y = y0;
        r.z = z0;
        return r;
    }
    Vec3 dup() {
        return create(x,y,z);
    }
    static Vec3 zero() {
        return create(0,0,0);
    }
    static Vec3 generate(float delegate() gen) {
        return create(gen(),gen(),gen());
    }

    // unary operators
    Vec3 opNeg() {
        return create(-x,-y,-z);
    }
    Vec3 opPos() {
        return dup();
    }
    // opCom
    // opPostInc
    // opPostDec
    // opCast

    // binary operators
    Vec3 opAdd(Vec3 v) {
        return create(x+v.x, y+v.y, z+v.z);
    }
    Vec3 opSub(Vec3 v) {
        return create(x-v.x, y-v.y, z-v.z);
    }
    Vec3 opMul(float v) {
        return create(x*v, y*v, z*v);
    }
    Vec3 opMul_r(float v) {
        return create(x*v, y*v, z*v);
    }
    Vec3 opDiv(float v) {
        return create(x/v, y/v, z/v);
    }
    // opMod
    // opAnd
    // opOr
    // opXor
    // opShl
    // opShr
    // opUShr
    // opCat
    bool opEquals(Vec3 v) {
        return x==v.x && y==v.y && z==v.z;
    }
    // opCmp
    // they are _VOID_ function
    void opAddAssign(Vec3 v) {
        x += v.x;
        y += v.y;
        z += v.z;
    }
    void opSubAssign(Vec3 v) {
        x -= v.x;
        y -= v.y;
        z -= v.z;
    }
    void opMulAssign(float v) {
        x *= v;
        y *= v;
        z *= v;
    }
    void opDivAssign(float v) {
        x /= v;
        y /= v;
        z /= v;
    }
    // opModAssign
    // opAndAssign
    // opOrAssign
    // opXorAssign
    // opShlAssign
    // opShrAssign
    // opUShrAssign
    // opCatAssign

    float opIndex(int i) { return _[i]; }
    float opIndexAssign(float v, int i) { return _[i] = v; }

    int opApply(int delegate(inout float) dg) {
        int result = 0;
        for (int i = 0; i < _.length; i++) {
            result = dg(_[i]);
            if (result) break;
        }
        return result;
    }
    int opApply(int delegate(inout float, inout int) dg) {
        int result = 0;
        for (int i = 0; i < _.length; i++) {
            result = dg(_[i], i);
            if (result) break;
        }
        return result;
    }

    float len2() {
        return x*x+y*y+z*z;
    }
    float len() {
        return sqrt(len2());
    }
    float dist2(Vec3 v) {
        return opSub(v).len2();
    }
    float dist(Vec3 v) {
        return opSub(v).len();
    }

    float dot(Vec3 v) {
        return x*v.x + y*v.y + z*v.z;
    }
    Vec3 cross(Vec3 v) {
        Vec3 r;
        r.x = y*v.z - z*v.y;
        r.y = z*v.x - x*v.z;
        r.z = x*v.y - y*v.x;
        return r;
    }

    void normalize() {
        float d = 1.0 / len();
        x *= d;
        y *= d;
        z *= d;
    }

    float angle(Vec3 v) {
        Vec3 c = cross(v);
        float s = c.len();
        return fabs(atan2(s, dot(v)));
    }

    char[] str() {
        return `(` ~ std.string.toString(x) ~ `,` ~
                 std.string.toString(y) ~ `,` ~
                 std.string.toString(z) ~ `)`;
    }
    char[] toString() {
        return str();
    }
}

Vec3 Vec(float x, float y, float z) {
    return Vec3.create(x, y, z);
}

unittest {
    bool approxEq(float x, float y) {
        return fabs(x-y) < 0.0001;
    }

    static Vec3 v1 = { x:1, y:1, z:1 };
    static Vec3 v2 = { x:2, y:3, z:4 };
    Vec3 v3 = v1 + v2;
    assert(v3.x == 3);
    assert(v3.y == 4);
    assert(v3.z == 5);
    assert(v3[X] == 3);
    assert(v3[Y] == 4);
    assert(v3[Z] == 5);
    assert(v1.dist2(v2) == 14);
    assert(v1.dot(v2) == 9);
    foreach (float v; v1) {
        assert(v == 1);
    }
    assert(approxEq(Vec3.create(1,0,0).angle(Vec3.create(0,1,1)), PI/2));
}
